(require 'eieio)
(require 'cl)

(when (not (fboundp 'set-process-query-on-exit-flag))
  (defalias 'set-process-query-on-exit-flag
	'process-kill-without-query))

(defclass srpc:active-node ()
  ((host            :initarg    :host
		            :initform   "localhost"
		            :protection :protected
					:documentation
					"接續先ホスト")
   (port            :initarg    :port
		            :initform   9895
		            :protection :protected
					:documentation
					"接續先ポート")
   (stream          :initform   nil
					:protection :protected
					:documentation
					"ソケット")
   (notify-callback :protection :protected
					:documentation
					"接續先ノードから notification が來た場合に呼ばれる關數")
   (request-id      :initform   0
					:protection :protected
					"次囘のリクエスト時に使用する ID")
   (inbound-buf     :initform ""
					:protection :protected
					:documentation
					"入力バッファ")
   (requests        :protection :protected
					:documentation
"リクエストを發行する際には、このハッシュテーブルにキー:
request-id, 値: nil として插入する。相手のノードから result が來た
時、對應するリクエストが存在するなら値を (result . 戻り値) また
は (error . エラー値) で置き換へるが、無ければ單に捨てる。")))

(defmethod initialize-instance ((this srpc:active-node) &optional fields)
  
  (oset this notify-callback
		(lambda (method arg)
		  (message "Notification from sRPC server: method %S: arg %S"
				   method arg)))

  (oset this requests
		(make-hash-table :test 'eq))

  (call-next-method))

(defmethod srpc:host ((this srpc:active-node))
  (with-slots (host) this
	host))

(defmethod srpc:port ((this srpc:active-node))
  (with-slots (port) this
	port))

(put 'srpc:failure
	 'error-conditions '(error srpc:failure))

(put 'srpc:failure
	 'error-message "An srpc call failure")

(defmethod srpc:connect ((this srpc:active-node))
  "相手のノードへ接續されてゐなければ接續する。"
  (block nil
	(with-slots (host port stream) this

	  ;; 既に繋がってゐるかどうかを確認
	  (if stream
		  (let ((status (process-status stream)))
			(if (or (eq status 'open)
					(eq status 'connect))
				;; 繋がってゐるか、繋げようとしてゐる最中なので今囘は何
				;; もしない。
				(return nil))))

	  ;; 接續
	  (setq stream (open-network-stream "sRPC" nil host port))
	  (oset this stream stream)
	  
	  (set-process-coding-system stream 'utf-8 'utf-8)
	  (set-process-query-on-exit-flag stream nil)

	  (lexical-let ((context this))
		(set-process-filter stream
							(lambda (proc str)
							  (srpc:process-inbound-chunk context str)))))))

(defmethod srpc:disconnect ((this srpc:active-node))
  "相手のノードへ接續されてゐれば切斷する。"
  (with-slots (stream) this
    (if stream
        (let ((status (process-status stream)))
          (if (or (eq status 'open)
                  (eq status 'connect))
              (progn
                (process-send-eof stream)
                (delete-process   stream)))))))

(defmethod srpc:send-s-exp ((this srpc:active-node) s-exp)
  "内部用"
  (with-slots (stream) this
	(srpc:connect this)

	(let ((line (apply #'concat
					   (mapcar (lambda (c)
								 (if (= c ?\n)
									 "\\n"
								   (char-to-string c)))
							   (format "%S" s-exp)))))
	  (process-send-string stream line)
	  (process-send-string stream "\n"))))

(defmethod srpc:notify ((this srpc:active-node) method &optional arg)
  "相手側ノードへ通知を送信する。"
  (srpc:send-s-exp this `((method . ,method)
						  (arg    . ,arg   ))))

(defmethod srpc:call ((this srpc:active-node) method &optional arg)
  "相手側ノードへ要求を送信し、返答を待つ。"
  (with-slots (stream request-id requests) this
	
	(let ((id request-id))
	  (setq request-id (1+ request-id))
	
	  (srpc:send-s-exp this `((method . ,method)
							  (id     . ,id)
							  (arg    . ,arg)))
 
	  (puthash id nil requests)
	  (while (null (gethash id requests 'no-such-request))
		;; 入力を待つ
		(accept-process-output stream))

	  (let* ((res   (gethash id requests 'no-such-request))
			 (type  (car res))
			 (value (cdr res)))

		(remhash id requests)
		
		(if (eq type 'result)
			;; 成功が返った
			value
		  
		  ;; 失敗が返った
		  (signal 'srpc:failure value))))))

(defmethod srpc:process-inbound-chunk ((this srpc:active-node) str)
  "内部用"
  (with-slots (inbound-buf) this
	
	;; バッファに追記
	(setq inbound-buf (concat inbound-buf str))
	
	;; 改行があれば、そこまでを切り出して處理
	(let ((newline-pos (search "\n" inbound-buf)))
	  (if newline-pos
		  
		  (let ((line (substring inbound-buf 0 newline-pos))
				outbound)
			
			(setq inbound-buf (substring inbound-buf (1+ newline-pos)))
			(setq outbound (srpc:process-inbound-line this line))

			(if outbound
				(srpc:send-s-exp this outbound)))))))

(defmethod srpc:process-inbound-line ((this srpc:active-node) line)
  "内部用"
  (block nil
	(with-slots (notify-callback requests) this
	  
	  (let (inbound)

		;; パースしてみる
		(condition-case e
			(setq inbound (read line))
		  (error
		   (return `((method . %parse-error)
					 (arg    . ,(format "I can't even parse your S-exp! [%s]" line))))))

		;; パース成功。次は受信したメッセージの種類を判別。
		(unless (listp inbound)
		  (return `((method . %syntax-error)
					(arg    . ,(format "Any messages have to be an alist! [%S]" inbound)))))

		(let ((method (assq 'method inbound))
			  (result (assq 'result inbound))
			  (error  (assq 'error  inbound))
			  (arg    (assq 'arg    inbound))
			  (id     (assq 'id     inbound)))
		
		  (cond ((and method id)
				 ;; 要求
				 (return `((error . ,(format "Sorry, this implementation doesn't support requests: %S"
											 inbound))
						   (id    . ,(cdr id)))))

				(method
				 ;; 通知
				 (funcall notify-callback
						  (cdr method)
						  (if arg
							  (cdr arg)
							nil))
				 (return nil))

				((and (or result error)
					  id)
				 ;; 應答
				 (let ((awaiting (gethash (cdr id) requests 'no-such-request)))
				   
				   (cond ((eq awaiting 'no-such-request)
						  ;; 待ってもゐない id に對して應答が來た。どちらかのバグ？
						  (message "Weird things have happened. A sRPC passive node gave me a response which (I think) I haven't sent any corresponding requests: %S"
								   inbound))

						 ((not (null awaiting))
						  ;; 同じ ID に對して二度以上の應答が來た。
						  (message "Weird things have happened. A sRPC passive node gave me two responses to a single request: %S"
								   inbound))

						 (t
						  (puthash (cdr id) (or result error) requests))))
				 (return nil))

				(t
				 ;; どれにも當て嵌らないのでエラーを返す
				 (return `((method . %syntax-error)
						   (arg    . ,(format "Unknown message type: %S" inbound)))))))))))

(provide 'srpc-active-node)